local mvec3_dir = mvector3.direction
local mvec3_dist_sq = mvector3.distance_sq
local mvec3_dot = mvector3.dot
local mvec3_cross = mvector3.cross
local mvec3_sub = mvector3.subtract
local mvec3_add = mvector3.add
local mvec3_mul = mvector3.multiply
local math_max = math.max
local tmp_vec1 = Vector3()
local tmp_vec2 = Vector3()
local up_offs = math.UP * 50

function SentryGunBrain:_get_tweak_data() --custom method
	if self._unit:weapon() then 
		return self._unit:weapon():_get_tweak_data()
	end
	return (self._name_id and tweak_data.weapon[self._name_id]) or tweak_data.weapon.sentry_gun
end

function SentryGunBrain:switch_on()
	if self._active or self._unit:character_damage():dead() then
		return
	end

	if self._unit:damage():has_sequence("laser_activate") then
		self._unit:damage():run_sequence_simple("laser_activate")
	end
	
--	self._unit:weapon():setup_criminal_laser()
--	self._unit:weapon():_set_sentry_laser(self._unit:weapon():_get_sentry_firemode())
--	local is_server = Network:is_server()


	self:set_active(true)
	self._ext_movement:switch_on()
	self._unit:set_slot(25)
	self._unit:base():register()
--[[
	if managers.groupai:state():all_criminals()[self._unit:key()] then
		managers.groupai:state():on_criminal_recovered(self._unit)
	end
--]]
--[[
	if is_server then
		PlayerMovement.set_attention_settings(self, {
			"sentry_gun_enemy_cbt"
		})
	end
	--]]
end

function SentryGunBrain:update(unit, t, dt)
	if self._unit:base():is_owner() or Network:is_server() then
		self:_upd_detection(t)
		self:_select_focus_attention(t)
	end
	if Network:is_server() then
		self:_upd_flash_grenade(t)
		self:_upd_go_idle(t)
	end

	self:_upd_fire(t)
end

local orig_target_on_sight = SentryGunBrain.is_target_on_sight
function SentryGunBrain:is_target_on_sight(my_pos, target_base_pos,...)
	if not self._unit:base()._owner_id then
		return orig_target_on_sight(self,my_pos,target_base_pos,...)
	end

	if not target_base_pos then
		return false
	end
	local td = self:_get_tweak_data()
	local fire_range = td.FIRE_RANGE
	local firemode = self._unit:weapon():_get_sentry_firemode()
	if firemode == "manual" then 
		fire_range = fire_range * 2
		
		--todo make sure this is the right kind of check 
		local los_ray = World:raycast("ray", my_pos, my_pos + (self._ext_movement:m_head_fwd() * fire_range), "ignore_unit", {self._unit}, "slot_mask",managers.slot:get_mask("bullet_impact_targets"))
		
		
		if los_ray then
			return true
		end

		
		local fire_angle = 5
		
		local vec_aim = self._ext_movement:m_head_fwd()
		mvector3.normalize(vec_aim)
		
		local angle_to_target = mvector3.angle(target_base_pos - my_pos,vec_aim)
--		Console:SetTrackerValue("trackere",string.format("%.2f",angle_to_target))
		if math.abs(angle_to_target) < fire_angle then 
			return true
		end
		
	else
		
		if firemode == "overwatch" then 
			fire_range = 10000
		end
		
		local fire_range_sq = fire_range * fire_range
		
		if fire_range_sq < mvec3_dist_sq(my_pos, target_base_pos) then
			return false
		end
	end

	local target_pos_same_height = mvector3.copy(target_base_pos)

	mvector3.set_z(target_pos_same_height, mvector3.z(my_pos))

	local dir_to_target = mvector3.copy(my_pos)
	mvec3_sub(dir_to_target, target_pos_same_height)
	mvector3.normalize(dir_to_target)

	local right_offset = Vector3()

	mvec3_cross(right_offset, dir_to_target, math.UP)

	local left_offset = mvector3.copy(right_offset)

	mvec3_mul(right_offset, self.attention_target_offset_hor)
	mvec3_mul(left_offset, -self.attention_target_offset_hor)
	mvector3.set_z(right_offset, mvector3.z(right_offset) + self.attention_target_offset_ver)
	mvector3.set_z(left_offset, mvector3.z(left_offset) + self.attention_target_offset_ver)

	local ignore_units = {
		self._unit
	}
	local offsets = {
		Vector3(0, 0, 0),
		right_offset,
		left_offset
	}

	for i, offset in ipairs(offsets) do
		local target_pos = mvector3.copy(target_base_pos)

		mvec3_add(target_pos, offset)

		local vis_ray = World:raycast("ray", my_pos, target_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision")

		if not vis_ray then
			return true
		end
	end

	return false
end


local orig_upd_fire = SentryGunBrain._upd_fire
function SentryGunBrain:_upd_fire(t,...)
	if self._unit:base():is_owner() then 
--		Console:SetTrackerValue("trackerb",Application:time())
		return orig_upd_fire(self,t,...)
	end
--	Console:SetTrackerValue("trackerb",string.format("%.2f",Application:time()))
	
	if self._ext_movement:is_activating() or self._ext_movement:is_inactivating() or self._idle then
		if self._firing then
			self:stop_autofire()
		end
		return
	end
	
	local td = self:_get_tweak_data()

	local attention = self._ext_movement:attention()

	if self._unit:weapon():out_of_ammo() then
		if self._unit:weapon():can_auto_reload() then
			if self._firing then
				self:stop_autofire()
			end

			if not self._ext_movement:rearming() then
				self._ext_movement:rearm()
			end
		elseif not self._unit:base():waiting_for_refill() then
			self:switch_off()
		end
	elseif self._ext_movement:rearming() then
		self._ext_movement:complete_rearming()
	elseif attention and attention.reaction and AIAttentionObject.REACT_SHOOT <= attention.reaction and not self._ext_movement:warming_up(t) and not self:_ignore_shield({
		self._unit
	}, self._ext_movement:m_head_pos(), self._attention_obj) then
		local expend_ammo = Network:is_server() and not is_criminal
		local damage_player = attention.unit:base() and attention.unit:base().is_local_player
		local my_pos = self._ext_movement:m_head_pos()
--		Console:SetFwdRayUnit(attention.unit)
		local target_pos = self:get_target_base_pos(attention)

		if not target_pos then
			self:stop_autofire()

			return
		end

--		Draw:brush(Color.red:with_alpha(0.3)):sphere(target_pos,150)
		
		if not self:is_target_on_sight(my_pos, target_pos) then
			self:stop_autofire()
			return
		end

		if self._firing then
			self._unit:weapon():trigger_held(false, expend_ammo, damage_player, attention.unit)
		else
			mvec3_dir(tmp_vec1, my_pos, target_pos)

			local max_dot = td.KEEP_FIRE_ANGLE
			max_dot = math.min(0.99, 1 - (1 - max_dot) * (self._shaprness_mul or 1))
			local this_dot = mvec3_dot(tmp_vec1, self._ext_movement:m_head_fwd())
			
--			Console:SetTrackerValue("trackerd",string.format("%.02f",this_dot))
			if max_dot < this_dot then
				self._unit:weapon():start_autofire()
				self._unit:weapon():trigger_held(false, expend_ammo, damage_player, attention.unit)

				self._firing = true
			end
			
		end
	elseif self._firing then
		self:stop_autofire()
	end
end

function SentryGunBrain:_ignore_shield(ignore_units, my_pos, attention)
	if self._unit:weapon():_get_ammo_type() == "ap" then
		return false
	end

	if attention and attention.unit then
		local hit_shield = World:raycast("ray", my_pos, attention.unit:position() + up_offs, "ignore_unit", ignore_units, "slot_mask", self._shield_check)

		if hit_shield then
			return true
		end
	end

	return false
end

function SentryGunBrain:_upd_detection(t)
	if self._ext_movement:is_activating() or self._ext_movement:is_inactivating() then
		return
	end

	if t < self._next_detection_upd_t then
		return
	end
	
	local td = self:_get_tweak_data()

	local delay = 1
	local my_SO_access_str = self._SO_access_str
	local my_SO_access = self._SO_access
	local my_tracker = self._unit:movement():nav_tracker()
	local chk_vis_func = my_tracker.check_visibility
	local detected_objects = self._detected_attention_objects
	local my_key = self._unit:key()
	local my_team = self._ext_movement:team()
	local my_pos = self._ext_movement:m_head_pos()
	local my_tracker = self._ext_movement:nav_tracker()
	local chk_vis_func = my_tracker.check_visibility
	local max_detection_range = td.DETECTION_RANGE
	local all_attention_objects = managers.groupai:state():get_AI_attention_objects_by_filter(my_SO_access_str, my_team)

	local firemode = self._unit:weapon():_get_sentry_firemode()

	local function _distance_chk(handler, settings, attention_pos)
		attention_pos = attention_pos or handler:get_detection_m_pos()
		local dis_sq = mvec3_dist_sq(my_pos, attention_pos)
		local max_dis = math.min(max_detection_range, settings.max_range or max_detection_range)

		if settings.detection and settings.detection.range_mul then
			max_dis = max_dis * settings.detection.range_mul
		end

		if dis_sq < max_dis * max_dis then
			return math.sqrt(dis_sq)
		end
		if firemode == "overwatch" then 
			return max_dis
		end
	end

	local ignore_units = {
		self._unit
	}

	local function _nearly_visible_chk(attention_info, detect_pos)
		local near_pos = tmp_vec1

		if attention_info.verified_dis < 2000 and math.abs(detect_pos.z - my_pos.z) < 300 then
			mvector3.set(near_pos, detect_pos)
			mvector3.set_z(near_pos, near_pos.z + 100)

			local near_vis_ray = World:raycast("ray", my_pos, near_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision", "report")

			if near_vis_ray then
				local side_vec = tmp_vec1

				mvector3.set(side_vec, detect_pos)
				mvec3_sub(side_vec, my_pos)
				mvec3_cross(side_vec, side_vec, math.UP)
				mvector3.set_length(side_vec, 150)
				mvector3.set(near_pos, detect_pos)
				mvec3_add(near_pos, side_vec)

				local near_vis_ray = World:raycast("ray", my_pos, near_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision", "report")

				if near_vis_ray then
					mvec3_mul(side_vec, -2)
					mvec3_add(near_pos, side_vec)

					near_vis_ray = World:raycast("ray", my_pos, near_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision", "report")
				end
			end

			if not near_vis_ray then
				attention_info.nearly_visible = true
				attention_info.last_verified_pos = mvector3.copy(near_pos)
			end
		end
	end

	for u_key, attention_info in pairs(all_attention_objects) do
		if u_key ~= my_key and not detected_objects[u_key] and (not attention_info.nav_tracker or chk_vis_func(my_tracker, attention_info.nav_tracker)) then
			local settings = attention_info.handler:get_attention(my_SO_access, AIAttentionObject.REACT_SUSPICIOUS, nil, my_team)

			if settings then
				local attention_pos = attention_info.handler:get_detection_m_pos()

				if _distance_chk(attention_info.handler, settings, attention_pos) then
					ignore_units[2] = attention_info.unit or nil
					local vis_ray = World:raycast("ray", my_pos, attention_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision")

					if not vis_ray or vis_ray.unit:key() == u_key then
						detected_objects[u_key] = CopLogicBase._create_detected_attention_object_data(t, self._unit, u_key, attention_info, settings)
					end
				end
			end
		end
	end

	local update_delay = 2

	for u_key, attention_info in pairs(detected_objects) do
		if t < attention_info.next_verify_t then
			update_delay = math.min(attention_info.next_verify_t - t, update_delay)
		else
			ignore_units[2] = attention_info.unit or nil
			attention_info.next_verify_t = t + (attention_info.identified and attention_info.verified and attention_info.settings.verification_interval or attention_info.settings.notice_interval or attention_info.settings.verification_interval)
			update_delay = math.min(update_delay, attention_info.settings.verification_interval)

			if not attention_info.identified then
				local health_ratio = self:_attention_health_ratio(attention_info)
				local objective = self:_attention_objective(attention_info)
				local noticable = nil
				local distance = _distance_chk(attention_info.handler, attention_info.settings, nil)
				local skip = objective == "surrender" or health_ratio <= 0

				if distance then
					local attention_pos = attention_info.handler:get_detection_m_pos()
					local vis_ray = World:raycast("ray", my_pos, attention_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision", "report")

					if not vis_ray or (firemode == "overwatch") then 
						noticable = true
					end
				end

				local delta_prog = nil
				local dt = t - attention_info.prev_notice_chk_t

				if noticable and not skip then
					local min_delay = td.DETECTION_DELAY[1][2]
					local max_delay = td.DETECTION_DELAY[2][2]
					local dis_ratio = (distance - td.DETECTION_DELAY[1][1]) / (td.DETECTION_DELAY[2][1] - td.DETECTION_DELAY[1][1])
					local dis_mul_mod = math.lerp(min_delay, max_delay, dis_ratio)
					local notice_delay_mul = attention_info.settings.notice_delay_mul or 1

					if attention_info.settings.detection and attention_info.settings.detection.delay_mul then
						notice_delay_mul = notice_delay_mul * attention_info.settings.detection.delay_mul
					end

					local notice_delay_modified = math.lerp(min_delay * notice_delay_mul, max_delay, dis_mul_mod)
					delta_prog = notice_delay_modified > 0 and dt / notice_delay_modified or 1
				else
					delta_prog = dt * -0.125
				end

				attention_info.notice_progress = attention_info.notice_progress + delta_prog

				if attention_info.notice_progress > 1 and not skip then
					attention_info.notice_progress = nil
					attention_info.prev_notice_chk_t = nil
					attention_info.identified = true
					attention_info.release_t = t + attention_info.settings.release_delay
					attention_info.identified_t = t
					noticable = true
				elseif attention_info.notice_progress < 0 or skip then
					self:_destroy_detected_attention_object_data(attention_info)

					noticable = false
				else
					noticable = attention_info.notice_progress
					attention_info.prev_notice_chk_t = t
				end

				if noticable ~= false and attention_info.settings.notice_clbk then
					attention_info.settings.notice_clbk(self._unit, noticable)
				end
			end

			if attention_info.identified then
				update_delay = math.min(update_delay, attention_info.settings.verification_interval)
				attention_info.nearly_visible = nil
				local verified, vis_ray = nil
				local attention_pos = attention_info.handler:get_detection_m_pos()
				local dis = mvector3.distance(my_pos, attention_info.m_head_pos)

				if dis < max_detection_range * 1.2 and (not attention_info.settings.max_range or dis < attention_info.settings.max_range * (attention_info.settings.detection and attention_info.settings.detection.range_mul or 1) * 1.2) then
					local detect_pos = nil

					if attention_info.is_husk_player and attention_info.unit:anim_data().crouch then
						detect_pos = tmp_vec1

						mvector3.set(detect_pos, attention_info.m_pos)
						mvec3_add(detect_pos, tweak_data.player.stances.default.crouched.head.translation)
					else
						detect_pos = attention_pos
					end

					vis_ray = World:raycast("ray", my_pos, detect_pos, "ignore_unit", ignore_units, "slot_mask", self._visibility_slotmask, "ray_type", "ai_vision")

					if not vis_ray then
						verified = true
					end

					attention_info.verified = verified
				end

				attention_info.dis = dis
				attention_info.vis_ray = vis_ray and vis_ray.dis or nil
				local is_downed = false

				if attention_info.unit:movement() and attention_info.unit:movement().downed then
					is_downed = attention_info.unit:movement():downed()
				end

				local is_ignored_target = self:_attention_health_ratio(attention_info) <= 0 or self:_attention_objective(attention_info) == "surrender" or is_downed

				if is_ignored_target then
					self:_destroy_detected_attention_object_data(attention_info)
				elseif verified and dis < td.FIRE_RANGE then
					attention_info.release_t = nil
					attention_info.verified_t = t

					mvector3.set(attention_info.verified_pos, attention_pos)

					attention_info.last_verified_pos = mvector3.copy(attention_pos)
					attention_info.verified_dis = dis
				elseif attention_info.has_team and my_team.foes[attention_info.unit:movement():team().id] then
					if attention_info.criminal_record and AIAttentionObject.REACT_COMBAT <= attention_info.settings.reaction then
						if dis > max_detection_range and mvector3.distance(attention_pos, attention_info.last_verified_pos or attention_info.criminal_record.pos) > 700 or max_detection_range < dis then
							self:_destroy_detected_attention_object_data(attention_info)
						else
							update_delay = math.min(0.2, update_delay)
							attention_info.verified_pos = mvector3.copy(attention_info.criminal_record.pos)
							attention_info.verified_dis = dis

							if vis_ray then
								_nearly_visible_chk(attention_info, attention_pos)
							end
						end
					elseif attention_info.release_t and attention_info.release_t < t then
						self:_destroy_detected_attention_object_data(attention_info)
					else
						attention_info.release_t = attention_info.release_t or t + attention_info.settings.release_delay
					end
				elseif attention_info.release_t and attention_info.release_t < t then
					self:_destroy_detected_attention_object_data(attention_info)
				else
					attention_info.release_t = attention_info.release_t or t + attention_info.settings.release_delay
				end
			end
		end
	end

	self._next_detection_upd_t = t + update_delay
end

local orig_select_attention = SentryGunBrain._select_focus_attention
function SentryGunBrain:_select_focus_attention(t,...)
	
	local weapon = self._unit:weapon()
	local firemode = weapon:_get_sentry_firemode()
	local ammotype = weapon:_get_ammo_type()
	
	if firemode ~= "overwatch" then 
		return orig_select_attention(self,t,...)
	end
	
	
	local current_focus = self._attention_obj
	local current_pos = self._ext_movement:m_head_pos()
	local current_fwd = nil

	if current_focus then
		current_fwd = tmp_vec2

		mvec3_dir(current_fwd, self._ext_movement:m_head_pos(), current_focus.m_head_pos)
	else
		current_fwd = self._ext_movement:m_head_fwd()
	end

	local max_dis = SentryControlMenu.tweakdata.OVERWATCH_DETECTION_RANGE

	local function _get_weight(attention_info)
		if not attention_info.identified then
			return
		end

		local total_weight = 1

		if attention_info.health_ratio and attention_info.unit:character_damage():health_ratio() <= 0 then
			return 0
		elseif attention_info.verified_t and t - attention_info.verified_t < 3 then
			local max_duration = 3
			local elapsed_t = t - attention_info.verified_t
			total_weight = total_weight * math.lerp(1, 0.6, elapsed_t / max_duration)
		else
			return 0
		end

		if attention_info.settings.weight_mul then
			total_weight = total_weight * attention_info.settings.weight_mul
		end

		local dis = mvec3_dir(tmp_vec1, current_pos, attention_info.m_head_pos)
		local dis_weight = math_max(0, (max_dis - dis) / max_dis)
		total_weight = total_weight * dis_weight
		local dot_weight = 1 + mvec3_dot(tmp_vec1, current_fwd)
		dot_weight = dot_weight * dot_weight * dot_weight
		total_weight = total_weight * dot_weight

		if self:_ignore_shield({
			self._unit
		}, current_pos, attention_info) then
			total_weight = total_weight * 0.01
		end

		return total_weight
	end

	local best_focus_attention, best_focus_weight = nil
	local best_focus_reaction = 0

	for u_key, attention_info in pairs(self._detected_attention_objects) do
		local weight = _get_weight(attention_info)
		if attention_info.unit then
			if attention_info.unit:base() and attention_info.unit:base()._tweak_table == "sniper" then 
				if (weight and (best_focus_reaction < attention_info.reaction or best_focus_reaction == attention_info.reaction and (not best_focus_weight or best_focus_weight < weight))) then
					best_focus_weight = weight
					best_focus_attention = attention_info
					best_focus_reaction = attention_info.reaction
				end
			end
		end
	end

	if current_focus ~= best_focus_attention then
		if best_focus_attention then
			best_focus_attention.settings = best_focus_attention.settings or {}
			best_focus_attention.settings.detection = best_focus_attention.settings.detection or {}
			best_focus_attention.settings.detection.range_mul = 100
			local attention_data = {
				unit = best_focus_attention.unit,
				u_key = best_focus_attention.u_key,
				handler = best_focus_attention.handler,
				reaction = best_focus_attention.reaction
			}

			self._ext_movement:set_attention(attention_data)
		else
			self._ext_movement:set_attention()
		end

		self._attention_obj = best_focus_attention
	end
end
