local tmp_rot1 = Rotation()


local mvec_to = Vector3()

local orig_update = SentryGunWeapon.update_laser
function SentryGunWeapon:update_laser(...)
	if not self._unit:base()._owner_id then 
		return orig_update(self,...)
	end
	
	if not self:has_laser() then
		return
	end

	if self._unit:character_damage():dead() then 
		self:set_laser_enabled(false)
		return
	elseif self._unit:movement():is_inactivating() or self._unit:movement():is_inactivated() then
	elseif self._unit:movement():is_activating() or self._unit:movement():rearming() then
--		self:set_laser_enabled("turret_module_rearming",true)
		return
	end
	self:_set_laser_state(true)
	if self._laser_unit then 
		self._laser_unit:base():set_on()
	end
end

function SentryGunWeapon:_get_tweak_data() --custom method
	if self._ammo_type == "taser" then 
		return tweak_data.weapon.sentry_taser
	elseif self._ammo_type == "ap" then 
		return tweak_data.weapon.sentry_ap
	elseif self._ammo_type == "he" then 
		return tweak_data.weapon.sentry_he
	else
		return (self._name_id and tweak_data.weapon[self._name_id]) or tweak_data.weapon.sentry_gun
	end
end

function SentryGunWeapon:_set_sentry_firemode(mode,play_sound)
	if self._sentry_firemode ~= mode then 
		if play_sound then 
			self._unit:sound_source():post_event("wp_sentrygun_swap_ammo")
		end
		self:stop_autofire()
		self:flip_fire_sound()
	end
	self:_set_sentry_laser(mode)
	self._sentry_firemode = mode
end

function SentryGunWeapon:_set_sentry_laser(mode)
    if self._laser_unit and self._laser_align then
		local color = Color(0,0,0,0)
		if mode == "manual" then
			color = SentryControlMenu.tweakdata.color_mode_manual
		elseif mode == "overwatch" then
			color = SentryControlMenu.tweakdata.color_mode_overwatch
		end
		if not self._unit:base():is_owner() then 
			self:set_laser_color(color:with_alpha(SentryControlMenu:GetTeammateSentryLaserAlpha()))
		else
			self:set_laser_color(color)
		end
	end
end

function SentryGunWeapon:_set_ammo_type(ammo_type,play_sound)
	if self._ammo_type ~= ammo_type then
		if play_sound then 
			self._unit:sound_source():post_event("wp_sentrygun_swap_ammo")
		end		
		self:stop_autofire()
		self:flip_fire_sound()
		self._use_armor_piecing = ammo_type == "ap"
	end
	
	self._ammo_type = ammo_type
end

function SentryGunWeapon:_get_ammo_type()
	return self._ammo_type
end

function SentryGunWeapon:_get_sentry_firemode()
	return self._sentry_firemode
end

function SentryGunWeapon:setup_criminal_laser() --creates the laser object, for player deployable sentries only
    if self._laser_unit and self._laser_align and self._unit:movement():team().id == "criminal1" then
        self._laser_unit:base():set_color(Color(0,0,0))
    elseif not alive(self._laser_unit) then
        self._laser_align = self._laser_align or self._unit:get_object(Idstring("fire"))
		if self._laser_align then 		
			local spawn_rot = self._laser_align:rotation()
			local spawn_pos = self._laser_align:position()
			self._laser_unit = World:spawn_unit(Idstring("units/payday2/weapons/wpn_npc_upg_fl_ass_smg_sho_peqbox/wpn_npc_upg_fl_ass_smg_sho_peqbox"), spawn_pos, spawn_rot)
			self._unit:link(self._laser_align:name(), self._laser_unit)
			self._laser_unit:base():set_npc()
			self._laser_unit:base():set_on()
			self._laser_unit:base():set_max_distace(10000)
			self._laser_unit:base():add_ray_ignore_unit(self._unit)
			self._laser_unit:base():set_color(Color(0,0,0))
			self._laser_unit:set_visible(false)
		end
    end
end

function SentryGunWeapon:set_laser_color(color)
	if alive(self._laser_unit) then 
		self._laser_unit:base():set_color(color)
	end
end



function SentryGunWeapon:init(unit)
	--todo make ws + worldhud here
	--[[
	if self._owner then 
		self._firemode = "taser" --tcd value
	else
		self._firemode = "basic"
	end
	--]]
	
--    self._laser_align = self._laser_align or unit:get_object(Idstring("fire"))
	if self._owner_id then 
		self._ammo_type = "basic"
		self._sentry_firemode = "normal"
	end
	
	self._unit = unit
	
	self._current_damage_mul = 1
	self._timer = TimerManager:game()
	self._character_slotmask = managers.slot:get_mask("raycastable_characters")
	self._next_fire_allowed = -1000
	self._obj_fire = self._unit:get_object(Idstring("a_detect"))
	self._effect_align = {
		self._unit:get_object(Idstring(self._muzzle_flash_parent or "fire")),
		self._unit:get_object(Idstring(self._muzzle_flash_parent or "fire"))
	}
	self._muzzle_flash_parent = nil

	if self._laser_align_name then
		self._laser_align = self._unit:get_object(Idstring(self._laser_align_name))
	end
	
	self._interleaving_fire = 1
	self._trail_effect_table = {
		effect = RaycastWeaponBase.TRAIL_EFFECT,
		position = Vector3(),
		normal = Vector3()
	}
	self._ammo_sync_resolution = 0.0625

	if Network:is_server() then
		self._ammo_total = 1
		self._ammo_max = self._ammo_total
		self._ammo_sync = 16
	else
		self._ammo_ratio = 1
	end

	self._spread_mul = 1
	self._use_armor_piercing = false
	self._slow_fire_rate = false
	self._fire_rate_reduction = 1
	self._name_id = self._unit:base():get_name_id()
	
	local my_tweak_data = self:_get_tweak_data()
	self._default_alert_size = my_tweak_data.alert_size
	self._from = Vector3()
	self._to = Vector3()
end

local orig__init = SentryGunWeapon._init
function SentryGunWeapon:_init(...)
	if not self._owner_id then 
		return orig__init(self,...)
	end
	self._ammo_type = "basic" --dummied for any non-owner clients

	self._laser_align = self._laser_align or self._unit:get_object(Idstring("fire"))
	
	self._name_id = self._unit:base():get_name_id()
	
	local my_tweak_data = self:_get_tweak_data()

--! changed from is_host check to is_owner check	
	self._bullet_slotmask = managers.slot:get_mask(self._unit:base():is_owner() and "bullet_impact_targets" or "bullet_blank_impact_targets")

	self._character_slotmask = managers.slot:get_mask("raycastable_characters")
	self._muzzle_effect = Idstring(my_tweak_data.muzzleflash or "effects/particles/test/muzzleflash_maingun")
	local muzzle_offset = Vector3()

	mvector3.set_static(muzzle_offset, 0, 10, 0)

	self._muzzle_effect_table = {
		{
			force_synch = false,
			effect = self._muzzle_effect,
			parent = self._effect_align[1]
		},
		{
			force_synch = false,
			effect = self._muzzle_effect,
			parent = self._effect_align[2]
		}
	}
	self._use_shell_ejection_effect = SystemInfo:platform() == Idstring("WIN32")

	if self._use_shell_ejection_effect then
		self._obj_shell_ejection = self._unit:get_object(Idstring("shell"))
		self._shell_ejection_effect = Idstring(tweak_data.weapon[self._name_id].shell_ejection or "effects/payday2/particles/weapons/shells/shell_556")
		self._shell_ejection_effect_table = {
			effect = self._shell_ejection_effect,
			parent = self._obj_shell_ejection
		}
	end

	self._damage = my_tweak_data.DAMAGE
	self._alert_events = {}
	self._alert_fires = {}
	self._suppression = my_tweak_data.SUPPRESSION
end

function SentryGunWeapon:setup(setup_data)
	if self._unit:base():is_owner() then 
		self._unit:base():_create_ws()
	end
	self:_init()
	
	--
	self:setup_criminal_laser()
	self:_set_sentry_laser(self._unit:weapon():_get_sentry_firemode())
	
	self._setup = setup_data
	self._default_alert_size = self._alert_size
	self._current_damage_mul = 1
	self._owner = setup_data.user_unit
	self._spread_mul = setup_data.spread_mul
	self._auto_reload = setup_data.auto_reload
	self._fire_rate_reduction = 1

	local td = self:_get_tweak_data() --changed reference to change alert size based on tweakdata rather than name_id
	if setup_data.alert_AI then
		self._alert_events = {}
		self._alert_size = self._alert_size or td.alert_size
		self._alert_fires = {}
	else
		self._alert_events = nil
		self._alert_size = nil
		self._alert_fires = nil
	end

	if setup_data.bullet_slotmask then
		self._bullet_slotmask = setup_data.bullet_slotmask
	end
	
end


function SentryGunWeapon:trigger_held(blanks, expend_ammo, shoot_player, target_unit)
	local fired = nil

	if self._next_fire_allowed <= self._timer:time() then
		fired = self:fire(blanks, false, shoot_player, target_unit)
		local td = self:_get_tweak_data()
		if fired then
			local fire_rate = td.auto.fire_rate * self._fire_rate_reduction
			self._next_fire_allowed = self._next_fire_allowed + fire_rate
			self._interleaving_fire = self._interleaving_fire == 1 and 2 or 1
		end
	end

	return fired
end

function SentryGunWeapon:fire(blanks, expend_ammo, shoot_player, target_unit)
	if expend_ammo then
		if self._ammo_total <= 0 then
			return
		end

		self:change_ammo(-1)
	end

	local fire_obj = self._effect_align[self._interleaving_fire]
	local from_pos = fire_obj:position()
	local direction = fire_obj:rotation():y()
	local td = self:_get_tweak_data()
	
	local SPREAD = td.SPREAD
	
	local td = self:_get_tweak_data()
	
	local firemode = self:_get_sentry_firemode()
	local ammotype = self:_get_ammo_type()
	
	if firemode == "manual" then 
		SPREAD = SPREAD * 2
	end
	
	if firemode ~= "overwatch" then 
		mvector3.spread(direction, td.SPREAD * self._spread_mul)
	end
	World:effect_manager():spawn(self._muzzle_effect_table[self._interleaving_fire])

	if self._use_shell_ejection_effect then
		World:effect_manager():spawn(self._shell_ejection_effect_table)
	end

	if self._unit:damage() and self._unit:damage():has_sequence("anim_fire_seq") then
		self._unit:damage():run_sequence_simple("anim_fire_seq")
	end

	local ray_res = self:_fire_raycast(from_pos, direction, shoot_player, target_unit)

	if self._alert_events and ray_res.rays then
		RaycastWeaponBase._check_alert(self, ray_res.rays, from_pos, direction, self._unit)
	end

	self._unit:movement():give_recoil()
	self._unit:event_listener():call("on_fire")

	return ray_res
end

function SentryGunWeapon:_fire_raycast(from_pos, direction, shoot_player, target_unit)
	local td = self:_get_tweak_data()
	local fire_range = td.FIRE_RANGE
	
	local firemode = self:_get_sentry_firemode()
	local ammotype = self:_get_ammo_type()
	
	local bullet_base = InstantBulletBase
	if ammotype == "taser" then 
		bullet_base = ElectricBulletBase
	elseif ammotype == "he" then 
		bullet_base = InstantExplosiveBulletBase
	end

	if firemode == "overwatch" then 
		fire_range = 100000
	end


	local result = {}
	local hit_unit, col_ray = nil

	mvector3.set(mvec_to, direction)
	mvector3.multiply(mvec_to, fire_range)
	mvector3.add(mvec_to, from_pos)

	self._from = from_pos
	self._to = mvec_to

	if not self._setup.ignore_units then
		return
	end

	if ammotype == "ap" then
		local col_rays = World:raycast_all("ray", from_pos, mvec_to, "slot_mask", self._bullet_slotmask, "ignore_unit", self._setup.ignore_units)
		col_ray = col_rays[1]

		if col_ray and col_ray.unit:in_slot(8) and alive(col_ray.unit:parent()) then
			col_ray = col_rays[2] or col_ray
		end
	else
		col_ray = World:raycast("ray", from_pos, mvec_to, "slot_mask", self._bullet_slotmask, "ignore_unit", self._setup.ignore_units)
	end

	local player_hit, player_ray_data = nil

	if shoot_player then
		player_hit, player_ray_data = RaycastWeaponBase.damage_player(self, col_ray, from_pos, direction)

		if player_hit then
			local damage = self:_apply_dmg_mul(self._damage, col_ray or player_ray_data, from_pos)

			bullet_base:on_hit_player(col_ray or player_ray_data, self._unit, self._unit, damage)
		end
	end

	if not player_hit and col_ray then
		local damage = self:_apply_dmg_mul(self._damage, col_ray, from_pos)
		hit_unit = bullet_base:on_collision(col_ray, self._unit, self._unit, damage)
	end

	if (not col_ray or col_ray.unit ~= target_unit) and target_unit and target_unit:character_damage() and target_unit:character_damage().build_suppression then
		target_unit:character_damage():build_suppression(self._suppression)
	end

	if not col_ray or col_ray.distance > 600 then
		self:_spawn_trail_effect(direction, col_ray)
	end

	result.hit_enemy = hit_unit

	if self._alert_events then
		result.rays = {
			col_ray
		}
	end

	return result
end

function SentryGunWeapon:auto_fire_start_event()
	local ammo_type = self:_get_ammo_type()
	if (ammo_type == "ap") or (ammo_type == "he") then
		return self._fire_start_snd_event_ap
	else
		return self._fire_start_snd_event
	end
end

function SentryGunWeapon:auto_fire_end_event()
	local ammo_type = self:_get_ammo_type()
	if (ammo_type == "ap") or (ammo_type == "he") then
		return self._fire_stop_snd_event_ap
	else
		return self._fire_stop_snd_event
	end
end


function SentryGunWeapon:_apply_dmg_mul(damage, col_ray, from_pos)
	local damage_out = damage * self._current_damage_mul
	
	local td = self:_get_tweak_data()
	
	if td.DAMAGE_MUL_RANGE then
		local ray_dis = col_ray.distance or mvector3.distance(from_pos, col_ray.position)
		local ranges = td.DAMAGE_MUL_RANGE
		local i_range = nil

		for test_i_range, range_data in ipairs(ranges) do
			if ray_dis < range_data[1] or test_i_range == #ranges then
				i_range = test_i_range

				break
			end
		end

		if i_range == 1 or ranges[i_range][1] < ray_dis then
			damage_out = damage_out * ranges[i_range][2]
		else
			local dis_lerp = (ray_dis - ranges[i_range - 1][1]) / (ranges[i_range][1] - ranges[i_range - 1][1])
			damage_out = damage_out * math.lerp(ranges[i_range - 1][2], ranges[i_range][2], dis_lerp)
		end
	end

	return damage_out
end
